@@ -3,6 +3,7 @@ package backendutil
 import (
 	"io"
 
+	"github.com/emersion/go-sasl"
 	"github.com/emersion/go-smtp"
 )
 
@@ -15,22 +16,12 @@ type TransformBackend struct {
 	TransformData func(r io.Reader) (io.Reader, error)
 }
 
-// Login implements the smtp.Backend interface.
-func (be *TransformBackend) Login(state *smtp.ConnectionState, username, password string) (smtp.Session, error) {
-	s, err := be.Backend.Login(state, username, password)
+func (be *TransformBackend) NewSession(c *smtp.Conn) (smtp.Session, error) {
+	sess, err := be.Backend.NewSession(c)
 	if err != nil {
 		return nil, err
 	}
-	return &transformSession{s, be}, nil
-}
-
-// AnonymousLogin implements the smtp.Backend interface.
-func (be *TransformBackend) AnonymousLogin(state *smtp.ConnectionState) (smtp.Session, error) {
-	s, err := be.Backend.AnonymousLogin(state)
-	if err != nil {
-		return nil, err
-	}
-	return &transformSession{s, be}, nil
+	return &transformSession{Session: sess, be: be}, nil
 }
 
 type transformSession struct {
@@ -43,7 +34,21 @@ func (s *transformSession) Reset() {
 	s.Session.Reset()
 }
 
-func (s *transformSession) Mail(from string, opts smtp.MailOptions) error {
+func (s *transformSession) AuthMechanisms() []string {
+	if authSession, ok := s.Session.(smtp.AuthSession); ok {
+		return authSession.AuthMechanisms()
+	}
+	return nil
+}
+
+func (s *transformSession) Auth(mech string) (sasl.Server, error) {
+	if authSession, ok := s.Session.(smtp.AuthSession); ok {
+		return authSession.Auth(mech)
+	}
+	return nil, smtp.ErrAuthUnsupported
+}
+
+func (s *transformSession) Mail(from string, opts *smtp.MailOptions) error {
 	if s.be.TransformMail != nil {
 		var err error
 		from, err = s.be.TransformMail(from)
@@ -54,7 +59,7 @@ func (s *transformSession) Mail(from string, opts smtp.MailOptions) error {
 	return s.Session.Mail(from, opts)
 }
 
-func (s *transformSession) Rcpt(to string) error {
+func (s *transformSession) Rcpt(to string, opts *smtp.RcptOptions) error {
 	if s.be.TransformRcpt != nil {
 		var err error
 		to, err = s.be.TransformRcpt(to)
@@ -62,7 +67,7 @@ func (s *transformSession) Rcpt(to string) error {
 			return err
 		}
 	}
-	return s.Session.Rcpt(to)
+	return s.Session.Rcpt(to, opts)
 }
 
 func (s *transformSession) Data(r io.Reader) error {
