@@ -10,6 +10,7 @@ import (
 	"strings"
 	"testing"
 
+	"github.com/emersion/go-sasl"
 	"github.com/emersion/go-smtp"
 	"github.com/emersion/go-smtp/backendutil"
 )
@@ -29,22 +30,7 @@ type backend struct {
 	userErr error
 }
 
-func (be *backend) Login(_ *smtp.ConnectionState, username, password string) (smtp.Session, error) {
-	if be.userErr != nil {
-		return &session{}, be.userErr
-	}
-
-	if username != "username" || password != "password" {
-		return nil, errors.New("Invalid username or password")
-	}
-	return &session{backend: be}, nil
-}
-
-func (be *backend) AnonymousLogin(_ *smtp.ConnectionState) (smtp.Session, error) {
-	if be.userErr != nil {
-		return &session{}, be.userErr
-	}
-
+func (be *backend) NewSession(c *smtp.Conn) (smtp.Session, error) {
 	return &session{backend: be, anonymous: true}, nil
 }
 
@@ -63,13 +49,33 @@ func (s *session) Logout() error {
 	return nil
 }
 
-func (s *session) Mail(from string, opts smtp.MailOptions) error {
+func (s *session) AuthMechanisms() []string {
+	return []string{sasl.Plain}
+}
+
+func (s *session) Auth(mech string) (sasl.Server, error) {
+	return sasl.NewPlainServer(func(identity, username, password string) error {
+		if identity != "" && identity != username {
+			return errors.New("Invalid identity")
+		}
+		if username != "username" || password != "password" {
+			return errors.New("Invalid username or password")
+		}
+		s.anonymous = false
+		return nil
+	}), nil
+}
+
+func (s *session) Mail(from string, opts *smtp.MailOptions) error {
+	if s.backend.userErr != nil {
+		return s.backend.userErr
+	}
 	s.Reset()
 	s.msg.From = from
 	return nil
 }
 
-func (s *session) Rcpt(to string) error {
+func (s *session) Rcpt(to string, opts *smtp.RcptOptions) error {
 	s.msg.To = append(s.msg.To, to)
 	return nil
 }
@@ -252,7 +258,7 @@ func TestServer(t *testing.T) {
 		t.Fatal("Invalid mail recipients:", msg.To)
 	}
 	// base64 of "Hey <3\n" (with actual newline)
-	if string(msg.Data) != "SGV5IDwzCg==" {
+	if string(msg.Data) != "SGV5IDwzDQo=" {
 		t.Fatal("Invalid mail data:", string(msg.Data))
 	}
 }
@@ -316,8 +322,8 @@ func TestServer_anonymousUserOK(t *testing.T) {
 	if len(msg.To) != 1 || msg.To[0] != "cm9vdEBnY2hxLmdvdi51aw==" {
 		t.Fatal("Invalid mail recipients:", msg.To)
 	}
-	// base64 of "Hey <3\n" (with actual newline)
-	if string(msg.Data) != "SGV5IDwzCg==" {
+	// base64 of "Hey <3\r\n" (with actual newline)
+	if string(msg.Data) != "SGV5IDwzDQo=" {
 		t.Fatal("Invalid mail data:", string(msg.Data))
 	}
 }
